
import time
from threading import Thread
from typing import List

try:
    from readchar import key as Key
    from readchar import readkey
except ModuleNotFoundError:
    print("Unable to find dependency module named 'readchar', install using 'pip install readchar'")
    exit(1)

import pybud.ansi as ansi
from pybud.drawing import ColoredString as CStr
from pybud.drawing import ColorType, Drawer
from .widgets import Widget
from ..deftypes import DEFAULT_BACKGROUND_COLOR

LAST_SHOWN_DRAWABLE = None
# ticks per second
TPS = 20

class Drawable():
    def __init__(self, ctype: ColorType = None):
        self.ctype = ctype
        self.width = None
        self.height = None
        self.background_color = DEFAULT_BACKGROUND_COLOR
        self.is_disabled = False
        self.closed = True
        self.drawer: Drawer = None
        self.drawing = False
        # self.last_update = 0
        self.tickupdate_thread: Thread = None
        self.tickupdate_started = False

    def onClose(self):
        self.close()

    def close(self):
        self.closed = True
        while self.tickupdate_thread.is_alive():
            time.sleep(0.01)
        self.tickupdate_started = False
        ansi.go_up(self.height)
        ansi.write(("\n" + " " * self.width) * (self.height))
        ansi.write(f"\033M" * self.height)
        ansi.write(f"\r" + " " * self.width)
        ansi.write(f"\033M\n")

    def update(self, key: str):
        if key == Key.CTRL_C:
            self.onClose()
        if key == Key.ESC:
            self.onClose()
        return key

    def doTickUpdates(self):
        t = time.time()
        while not self.closed:
            time.sleep(max(0, 1/TPS - (time.time() - t)))
            self.update("UPDATE")
            t = time.time()

    def doKeyUpdates(self):
        while not self.closed:
            try:
                key = readkey()
            except KeyboardInterrupt:
                key = Key.CTRL_C
            self.update(key)

    def show(self):
        global LAST_SHOWN_DRAWABLE
        LAST_SHOWN_DRAWABLE = self

        self.closed = False

        ansi.write("\n" * self.height)
        self.draw()

        if not self.tickupdate_started:
            self.tickupdate_thread = Thread(target=self.doTickUpdates)
            self.tickupdate_thread.start()
            self.tickupdate_started = True
        self.doKeyUpdates()

    def get_drawer(self):
        return Drawer(size=(self.width, self.height), background_color=self.background_color)

    def draw(self):
        if not self.closed:
            self.drawing = True
            self.drawer = self.get_drawer()


class DialogBase(Drawable):
    def __init__(self, width: int, ctype: ColorType = None, background_color: tuple[int, int, int] = None):
        super().__init__(ctype)
        self.width = width
        self.background_color = background_color
        self.widgets: List[Widget] = []
        self.set_active_widget(0)
        self.result = None
        self.tick = 0
        self.totw = 0

    def update_height(self):
        max_height = 0
        for w in self.widgets:
            widget_height = w.pos.getY() + w.size.getHeight()
            if widget_height > max_height:
                max_height = widget_height

        # set the height to one line more than the end of most buttom added dialog
        self.height = max_height + 1

    def add_widget(self, w: Widget):
        w.on_add(self)
        self.widgets.append(w)
        # self.height += w.size.getHeight()
        self.update_height()
        self.totw = self.get_total_selectable_widgets()

    def get_total_selectable_widgets(self):
        i = 0
        for w in self.widgets:
            if w.selectable:
                i += 1
        return i

    def get_active_widget(self, return_i: bool = False):
        i = 0
        w = None

        if self.totw == 0:
            if return_i:
                return w, i
            else:
                return w
        
        for w in self.widgets:
            if not w.selectable:
                continue
            if w.is_disabled:
                i += 1
                continue
            break

        if return_i:
            return w, i
        else:
            return w

    def set_active_widget(self, __i: int):
        i = 0
        for w in self.widgets:
            if not w.selectable:
                continue

            if i == __i:
                w.is_disabled = False
            else:
                w.is_disabled = True
            i += 1

    def update(self, key, draw=True):
        key = super().update(key)

        w, i = self.get_active_widget(True)
        if w is not None:
            key = w.update(key)
            self.result = w.result

        if key == Key.TAB:
            self.set_active_widget((i + 1) % self.totw)

        if key == "UPDATE":
            self.tick += 1

        if draw:
            self.draw()
        else:
            return key

    def draw_widgets(self, drawer: Drawer):
        active_w = self.get_active_widget()
        for w in self.widgets:
            border = w is active_w
            w_render = w.get_render()
            drawer.place_drawer(
                w_render, (w.pos.getX(), w.pos.getY()), borderless=not border)
        return drawer

    def render(self):
        self.drawer = self.draw_widgets(self.drawer)

    def draw(self):
        if self.closed or self.drawing:
            return
        super().draw()
        self.render()
        ansi.go_up(self.height)
        ansi.write(self.drawer.tostring(self.ctype) + "\n")
        ansi.flush()
        # time.sleep(0.01)
        self.drawing = False

    def show(self):
        super().show()
        return self.result

    def __str__(self):
        s = "+ " + self.__class__.__name__ + ":\n"
        for w in self.widgets:
            s += "|   " + w.name + "\n"
        return s


class AutoDialog(DialogBase):
    def __init__(self, width: int, ctype: ColorType = None, background_color: tuple[int, int, int] = None, mode: str = "v"):
        super().__init__(width, ctype)
        assert mode in ["v", "iv", "h", "ih"], f"Unknown mode \"{mode}\"!"
        self.mode = mode.lower()
        self.background_color = background_color

    def update(self, key):
        key = super().update(key, draw=False)
        if key == None:
            self.draw()
            return

        if self.totw != 0:
            w, i = self.get_active_widget(True)
            av = i
            if self.mode == "v":
                if key == Key.UP:
                    av = (i - 1) % self.totw 
                elif key == Key.DOWN:
                    av = (i + 1) % self.totw
            elif self.mode == "iv":
                if key == Key.DOWN:
                    av = (i + 1) % self.totw
                elif key == Key.UP:
                    av = (i + 1) % self.totw
            elif self.mode == "h":
                if key == Key.LEFT:
                    av = (i - 1) % self.totw
                elif key == Key.RIGHT:
                    av = (i + 1) % self.totw
            elif self.mode == "ih":
                if key == Key.RIGHT:
                    av = (i + 1) % self.totw
                elif key == Key.LEFT:
                    av = (i - 1) % self.totw
            else:
                raise ValueError(f"Unknown mode \"{self.mode}\"!")
            self.set_active_widget(av)
        
        # draw after each update
        self.draw()

    def render(self):
        super().render()
        animation = "▁▂▃▄▅▆▆▅▄▃▂▁▂ "

        def get_admination(tick, n=3):
            rt = tick
            return animation[rt % (len(animation)-n):rt % (len(animation)-n)+n]
        self.drawer.place(CStr(get_admination(self.tick)), pos=(1, 0))

    def show(self):
        super().show()
        return self.result



DEFAULT_BACKGROUND_COLOR = None  # (90, 90, 255)


class Point:
    def __init__(self, x: int = 0, y: int = 0) -> None:
        assert x >= 0
        assert y >= 0
        self.x = x
        self.y = y

    def getX(self):
        return self.x

    def getY(self):
        return self.y

    def distance_squared(self, other):
        if isinstance(other, Point):
            return (self.x - other.x) ** 2 + (self.y - other.y) ** 2
        elif isinstance(other, tuple) or isinstance(other, list):
            return (self.x - other[0]) ** 2 + (self.y - other[1]) ** 2
        else:
            raise NotImplementedError()

    def distance(self, other):
        return self.distance_squared(other) ** 0.5

    def __add__(self, other):
        if isinstance(other, Point):
            return Point(self.x + other.x, self.y + other.y)
        elif isinstance(other, tuple) or isinstance(other, list):
            return Point(self.x + other[0], self.y + other[1])
        else:
            raise NotImplementedError()

    def __iadd__(self, p2: 'Point'):
        return self + p2


class Size:
    def __init__(self, w: int = 0, h: int = 0) -> None:
        assert w >= 0
        assert h >= 0
        self.w = w
        self.h = h

    def getWidth(self) -> int:
        return self.w

    def getHeight(self) -> int:
        return self.h

    def getArea(self):
        return self.getWidth() * self.getHeight()

    def __add__(self, other):
        if isinstance(other, Size):
            return Point(self.getWidth() + other.getWidth(), self.getHeight() + other.getHeight())
        elif isinstance(other, tuple) or isinstance(other, list):
            return Point(self.getWidth() + other[0], self.getHeight() + other[1])
        else:
            raise NotImplementedError()

    def __iadd__(self, other):
        return self + other



from typing import SupportsIndex, overload

import pybud.ansi as ansi
from pybud.ansi import CR, AnsiColor, ColorType

write_to_file = False

if write_to_file:
    f = open("output.ansi", "w", encoding="utf-8")
    ansi.write = lambda x: f.write(x)
    ansi.writeln = lambda x: f.write(x + "\n")
    ansi.flush = lambda: f.flush()

# ansilen = lambda x: len(ansi.remove_ansi(x))

def get_rgb_if_exists(color: AnsiColor = None):
    return color.rgb if color is not None else None

class ColoredString():
    def __init__(self, string: str, forecolor: tuple[int, int, int] = None, backcolor: tuple[int, int, int] = None, ctype: ColorType = ColorType.LEGACY):
        self.forecolor = AnsiColor(
            rgb=forecolor) if forecolor is not None else None
        self.backcolor = AnsiColor(
            rgb=backcolor, foreground=False) if backcolor is not None else None
        self.str = string
        self.ctype = ctype

    def __len__(self) -> int:
        return len(self.str)

    def tostring(self, ctype: ColorType = None, ignore_background: bool = False):
        __ctype = ctype if ctype is not None else self.ctype
        color = ""
        
        if (not ignore_background) and (self.backcolor is not None):
            color += self.backcolor.get(ctype=__ctype)

        if self.forecolor is not None:
            color += self.forecolor.get(ctype=__ctype)

        return (color + self.str + CR)

    def __str__(self) -> str:
        return self.tostring(self.ctype)

    def __add__(self, other):
        if isinstance(other, str):
            other_ = ColoredString(other)
            return self + other_

        if isinstance(other, ColoredStringList):
            return ColoredStringList.__add__(ColoredStringList([self]), other)

        if self.forecolor == other.forecolor and self.backcolor == other.backcolor:
            return ColoredString(
                string=self.str + other.str,
                forecolor=get_rgb_if_exists(self.forecolor),
                backcolor=get_rgb_if_exists(self.backcolor)
            )
        else:
            return ColoredStringList([self, other])

    def __mul__ (self, other):
        if isinstance(other, int):
            return ColoredString(
                string=self.str * other,
                forecolor=get_rgb_if_exists(self.forecolor),
                backcolor=get_rgb_if_exists(self.backcolor)
            )
        else:
            raise NotImplementedError()

    def __iadd__(self, other):
        return self + other

    @overload
    def __getitem__(self, __p: SupportsIndex): ...

    @overload
    def __getitem__(self, __s: slice): ...

    def __getitem__(self, __p):
        if isinstance(__p, SupportsIndex):
            return ColoredString(
                string=self.str[__p],
                forecolor=get_rgb_if_exists(self.forecolor),
                backcolor=get_rgb_if_exists(self.backcolor)
            )
        elif isinstance(__p, slice):
            return ColoredString(
                string=self.str[__p.start:__p.stop:__p.step],
                forecolor=get_rgb_if_exists(self.forecolor),
                backcolor=get_rgb_if_exists(self.backcolor)
            )

    @overload
    def __setitem__(self, __p: SupportsIndex, __o): ...

    @overload
    def __setitem__(self, __p: slice, __o): ...

    # @overload
    def __setitem__(self, __p, __o):
        if isinstance(__p, SupportsIndex):
            if isinstance(__o, str):
                self.str[__p] = __o
            elif isinstance(__o, ColoredString):
                if len(__o.str) != 1:
                    raise ValueError("Shape mismatch! can not set an index of ColoredString to multiple characters.")
                else:
                    if self.forecolor == __o.forecolor and self.backcolor == __o.backcolor:
                        self.str[__p] = __o.str
                    elif len(self.str) == 1:
                        self.str = __o.str
                    else:
                        raise NotImplementedError("Setting an index of a `ColoredString` to another color is not supported!")
        elif isinstance(__p, slice):
            # the code below is comented because if there's a shape mismatch, this
            # will raise a ValueError for setting self.str anyways...
            # if len(__o.str) != len(self.strp[__p.start:__p.stop:__p.step]):
            #    raise ValueError("Shape mismatch! make sure to set to correct lenght.")

            if isinstance(__o, str):
                self.str[__p.start:__p.stop:__p.step] = __o
            elif isinstance(__o, ColoredString):
                if self.forecolor == __o.forecolor and self.backcolor == __o.backcolor:
                    self.str[__p.start:__p.stop:__p.step] = __o
                elif len(self.str) == len(__o.str):
                    self.str = __o.str
                else:
                    assert __p.step == 1, "step is not supported when differentiating colors!"
                    if self.str[:__p.start] != "":
                        cstr_pre = ColoredString(
                            string=self.str[:__p.start],
                            forecolor=get_rgb_if_exists(self.forecolor),
                            backcolor=get_rgb_if_exists(self.backcolor)
                        )
                    else:
                        cstr_pre = None
                    cstr_center = ColoredString(
                        string=__o.str,
                        forecolor=get_rgb_if_exists(__o.forecolor),
                        backcolor=get_rgb_if_exists(__o.backcolor)
                    )
                    if self.str[__p.stop+len(__o.str):] != "":
                        cstr_post = ColoredString(
                            string=self.str[__p.stop+len(__o.str):],
                            forecolor=get_rgb_if_exists(self.forecolor),
                            backcolor=get_rgb_if_exists(self.backcolor)
                        )
                    else:
                        cstr_post = None

                    # return the simplest representation
                    if cstr_post is None and cstr_pre is None:
                        return cstr_center
                    elif cstr_post is None:
                        return ColoredStringList([cstr_pre, cstr_center])
                    elif cstr_pre is None:
                        return ColoredStringList([cstr_center, cstr_post])
                    else:
                        return ColoredStringList([cstr_pre, cstr_center, cstr_post])

    @overload
    def __delitem__(self, __i: SupportsIndex): ...

    @overload
    def __delitem__(self, __s: slice): ...

    def __delitem__(self, __p):
        if isinstance(__p, SupportsIndex):
            del self.str[__p]
        elif isinstance(__p, slice):
            del self.str[__p.start:__p.stop:__p.step]

    # def __add__(self, other:str):
    #    if self.forecolor == None and self.backcolor == None:
    #        return self.str + other
    #    else:
    #        other_ = ColoredString(string=self.str + other)
    #        return ColoredStringList([self, other_])
    def __iadd__(self, other):
        return self + other

# modify built-in string class
# from forbiddenfruit import curse

# def curse_str():
#    def __add__(self, other: ColoredString):
#        if isinstance(other, ColoredString):
#            return ColoredString(self) + other
#        elif isinstance(other, ColoredStringList):
#            return ColoredString(self) + other
#        else:
#            return super(str, self).__add__(other)
#    curse(str, "__add__", __add__)
# curse_str()

class ColoredStringList():
    def __init__(self, strings: ColoredString, ctype: ColorType = ColorType.LEGACY):
        self.strs: list[ColoredString] = strings
        self.ctype = ctype

    def copy(self):
        return ColoredStringList(list(self.strs))

    def tostring(self, ctype: ColorType = None, ignore_background: bool = False):
        result = ""
        fc_c = None
        bc_c = None
        for cstr in self.strs:
            fc_now = cstr.forecolor
            bc_now = cstr.backcolor

            if fc_now != fc_c:
                if fc_now is not None:
                    result += fc_now.get(ctype)
                else:
                    result += ansi.ForeColors.Default
                fc_c = fc_now

            if (not ignore_background) and (bc_now != bc_c):
                if bc_now is not None:
                    result += bc_now.get(ctype)
                else:
                    result += ansi.BackColors.Default
                fc_c = fc_now

            result += cstr.str
        return result.replace("\n", ansi.CR + "\n")

    def __str__(self) -> str:
        return self.tostring(self.ctype)

    def __len__(self) -> int:
        return sum(map(len, self.strs))

    @overload
    def __add__(self, other: 'ColoredStringList'): ...

    @overload
    def __add__(self, other: ColoredString): ...

    @overload
    def __add__(self, other: str): ...

    def __add__(self, other):
        if isinstance(other, ColoredStringList):
            return ColoredStringList(self.strs + other.strs)
        elif isinstance(other, ColoredString):
            # return ColoredStringList(self.strs + [other]) (is not optimized)

            # if self.strs[-1] + other can be optimized to a single ColoredString object
            # this will return a ColoredString, else it will return a ColoredStringList
            # the value then will be added to self.strs[:-1]
            new_last = self.strs[-1] + other
            if isinstance(new_last, ColoredString):
                return ColoredStringList(self.strs[:-1] + [new_last])
            elif isinstance(new_last, ColoredStringList):
                return ColoredStringList(self.strs[:-1] + new_last.strs)
        elif isinstance(other, str):
            return self + ColoredString(string=other)
        else:
            raise NotImplementedError(
                f"`ColoredStringList` Only supports to be added to a \"ColoredString\", \"ColoredStringList\" or \"str\", but got \"{type(other)}\"")

    def __iadd__(self, other):
        return self + other

    @overload
    def __getitem__(self, __p: SupportsIndex): ...

    @overload
    def __getitem__(self, __s: slice): ...

    def __getitem__(self, __p):
        if isinstance(__p, SupportsIndex):
            p = 0
            for cstr in self.strs:
                # if cstr is empty continue (this should never happen but we want to be optimized)
                if len(cstr) == 0:
                    continue
                if (__p - p - len(cstr)) < 0:
                    r = cstr[__p-p]
                    return r
                else:
                    p += len(cstr)
            raise IndexError("Index out of range!")
        elif isinstance(__p, slice):
            # TODO: optimize this
            st = 0 if __p.start is None else __p.start
            et = len(self) if __p.stop is None else __p.stop
            s = 1 if __p.step is None else __p.step
            r = ColoredString("")
            for i in range(st, et, s):
                r += self[i]
            return r


def assign_to_cstrlist(text: ColoredString, __list: list[ColoredString], __i: SupportsIndex):
    l = __list.copy()

    text_len = len(text)
    idx_c = 0
    text_pos_c = 0
    for i, c in enumerate(l):
        if (idx_c + len(c)) > __i:
            # sel_id = i
            el_len = len(l[i])
            relative_pos = __i-idx_c+text_pos_c
            write_len = min(el_len - relative_pos, text_len-text_pos_c)
            backcolor_overwrite = get_rgb_if_exists(text.backcolor)
            if backcolor_overwrite is None:
                backcolor_overwrite = get_rgb_if_exists(l[i].backcolor)
            text_overwrite = ColoredString(
                string=text[text_pos_c:text_pos_c+write_len].str,
                forecolor=get_rgb_if_exists(text.forecolor),
                backcolor=backcolor_overwrite
            )
            if write_len == el_len:
                l[i] = text_overwrite
            else:
                l[i] = [l[i][:relative_pos], text_overwrite,
                        l[i][relative_pos+write_len:]]
            if ((text_pos_c+write_len) >= text_len):
                break
            text_pos_c += write_len
        idx_c += len(c)
    l_out = []
    for el in l:
        if isinstance(el, list):
            l_out += [__el for __el in el if __el != ""]
        else:
            l_out.append(el)
    return l_out


def write_to_cstrlist(text: str, __list: list[ColoredString], __i: SupportsIndex):
    l = __list.copy()
    text_len = len(text)
    idx_c = 0
    text_pos_c = 0
    for i, c in enumerate(l):
        if (idx_c + len(c)) > __i:
            # sel_id = i
            el_len = len(l[i])
            relative_pos = __i-idx_c+text_pos_c
            write_len = min(el_len - relative_pos, text_len-text_pos_c)
            text_overwrite = ColoredString(
                string=text[text_pos_c:text_pos_c+write_len],
                backcolor=l[i].backcolor.rgb if l[i].backcolor else None
            )
            if write_len == el_len:
                l[i] = text_overwrite
            else:
                l[i] = [l[i][:relative_pos], text_overwrite,
                        l[i][relative_pos+write_len:]]
            if ((text_pos_c+write_len) >= text_len):
                break
            text_pos_c += write_len
        idx_c += len(c)
    l_out = []
    for el in l:
        if isinstance(el, list):
            l_out += [__el for __el in el if __el != ""]
        else:
            l_out.append(el)
    return l_out


class Drawer(object):
    def __init__(self, size: tuple[int, int], background_color: tuple[int, int, int] = None):
        self.w = size[0]
        self.h = size[1]

        self.plane: list = [ColoredString(
            " " * self.w, backcolor=background_color) for i in range(self.h)]

    def __str__(self):
        return self.tostring()

    def tostring(self, ctype: ColorType = None, ignore_background: bool = False):
        result = ColoredString("")
        for i, line in enumerate(self.plane):
            if (i < (len(self.plane)-1)):
                result += line + "\n"
            else:
                result += line
        return result.tostring(ctype, ignore_background) + CR

    def write_text(self, text: str, pos: list[int, int], color: list[int, int, int] = None):
        assert isinstance(text, str)
        idx = pos[0]
        ln_idx = pos[1]

        if (idx >= self.w):
            return
        if (ln_idx < 0 or ln_idx >= self.h):
            return

        # fix text clipping out of frame
        if (idx < 0):
            text = text[-idx:]
            idx = 0
        text = text[:self.w - idx]

        line = self.plane[ln_idx]
        if isinstance(line, ColoredString):
            line_new = line[:idx] + ColoredString(
                string=text,
                forecolor=color,
                backcolor=get_rgb_if_exists(line.backcolor)
            ) + line[idx+len(text):]
            self.plane[ln_idx] = line_new
        elif isinstance(line, ColoredStringList):
            new_strs = write_to_cstrlist(text, line.strs, idx)
            new_line = ColoredStringList(new_strs)
            self.plane[ln_idx] = new_line

    def place(self, text, pos: list[int, int]):
        idx = pos[0]
        ln_idx = pos[1]
        if (idx >= self.w):
            return
        if (ln_idx < 0 or ln_idx >= self.h):
            return
        line = self.plane[ln_idx]
        if isinstance(text, ColoredStringList):
            idx_c = idx
            for cstr in text.strs:
                self.place(cstr, (idx_c, ln_idx))
                idx_c += len(cstr)
            return
        if len(text) == 0:
            return

        # fix text clipping out of frame
        if (idx < 0):
            text = text[-idx:]
            idx = 0
        text = text[:self.w - idx]
    
        if isinstance(line, ColoredString):
            backcolor_overwrite = get_rgb_if_exists(text.backcolor)
            if backcolor_overwrite is None:
                backcolor_overwrite = get_rgb_if_exists(line.backcolor)
            new_line = line[:idx] + ColoredString(
                string=text.str,
                forecolor=get_rgb_if_exists(text.forecolor),
                backcolor=backcolor_overwrite
            ) + line[idx+len(text):]
            self.plane[ln_idx] = new_line
        elif isinstance(line, ColoredStringList):
            new_strs = assign_to_cstrlist(text, line.strs, idx)
            new_line = ColoredStringList(new_strs)
            self.plane[ln_idx] = new_line

    def center_write(self, text, ln_idx: int):
        idx = (self.w - len(text))//2
        self.write_text(text, pos=(ln_idx, idx))

    def center_place(self, text, ln_idx: int):
        idx = (self.w - len(text))//2
        self.place(text, pos=(idx, ln_idx))
    
    def place_drawer(self, other_drawer: 'Drawer', pos: list[int, int], borderless: bool = True, shadow: bool = False):
        idx = pos[0]
        ln_idx = pos[1]
        if (idx >= self.w):
            return
        if (ln_idx < 0 or ln_idx >= self.h):
            return
        
        if not borderless:
            # shadow character
            sc = ColoredString("\\", forecolor=(180, 180, 180)) if shadow else ColoredString(" ")
            # place window with border
            self.place(ColoredString("┌") + ColoredString("─" * other_drawer.w) + ColoredString("┐"), (idx-1, ln_idx - 1))
            i = 0
            for i, cstr in enumerate(other_drawer.plane):
                self.place(sc + ColoredString("│") + cstr + ColoredString("│"), (idx-2, i + ln_idx))
            self.place(sc + ColoredString("└") + ColoredString("─" * other_drawer.w) + ColoredString("┘"), (idx-2, i + ln_idx + 1))
            self.place(sc * (other_drawer.w + 3), (idx-2, i + ln_idx + 2))
            return
        else:
            for i, cstr in enumerate(other_drawer.plane):
                self.place(cstr, (idx, i + ln_idx))
            return
    
    def center_place_drawer(self, other_drawer: 'Drawer', ln_idx: int, borderless: bool = True):
        idx = (self.w - other_drawer.w)//2
        self.place_drawer(other_drawer, pos=(idx, ln_idx), borderless = borderless)

if __name__ == "__main__":
    #result = "test" + (ColoredString("will this work?") + ColoredString("will this work?", forecolor=(0, 255, 0)))
    #print(result)

    # test = ColoredString("test ", forecolor=(0, 100, 200)) + ColoredString("for ", forecolor=(0, 100, 200)) + ColoredString("colors", forecolor=(0, 100, 200), backcolor=(200, 100, 200))
    tsize = ansi.get_size()
    screen = Drawer(size=(tsize.columns, tsize.lines))
    drawer = Drawer(size=(31, 5))
    drawer.write_text("Hello, World!", ((drawer.w-13)//2, 2))
    # painter.place(ColoredString("Hello, World!", forecolor=(200, 100, 0)), (8, 5))
    welcome_text = ColoredString("-"*((drawer.w-15)//2)+"[", forecolor=(0, 255, 0)) + ColoredString(
        "Welcome", forecolor=(255, 255, 0)) + ColoredString("]"+"-"*((drawer.w-15)//2), forecolor=(0, 255, 0))
    test_text = ColoredString("-"*((drawer.w-15)//2)+"[", forecolor=(0, 255, 0)) + ColoredString(
        "TEST", forecolor=(255, 255, 0)) + ColoredString("]"+"-"*((drawer.w-15)//2), forecolor=(0, 0, 0))
    drawer.center_place(welcome_text, ln_idx=1)
    drawer.place(test_text, pos=(0, 1))
    # print(type(drawer))
    screen.center_place_drawer(drawer, 10, borderless = False)
    ansi.write(screen.tostring(ctype=ColorType.TRUECOLOR))
    ansi.flush()
    input()




import types

from pybud.drawing import ColoredString as CStr
from pybud.drawing import ColoredStringList as CStrList
from pybud.drawing import ColorType, Drawer
from readchar import key as Key

from ..deftypes import Point, Size, DEFAULT_BACKGROUND_COLOR

def default(d: dict, k:str, default):
    if k in d.keys():
        return d[k]
    else:
        return default

class WidgetBase():
    def __init__(self, size: Size = None, pos: Point = Point(0, 0), **kwargs):
        # set variabled
        self.ctype = default(kwargs, "ctype", ColorType.LEGACY)
        self.size = Size(10, 20) if size is None else size
        self.pos = Point(0, 0) if pos is None else pos
        # defaults
        self.background_color = DEFAULT_BACKGROUND_COLOR
        self.name = default(kwargs, "name", "__name__")
        self.parent = None
        
        self.result = None

    def get_name(self):
        name =  self.__class__.__name__
        c = 0
        for w in self.parent.widgets:
            if isinstance(w, self.__class__):
                c += 1
        return name + "_" + str(c+1)

    
    def on_add(self, parent):
        self.parent = parent
        if self.name == "__name__":
            self.name = self.get_name()

    def on_interrupt(self):
        pass

    def get_drawer(self):
        return Drawer(size=(self.size.getWidth(), self.size.getHeight()), background_color=self.background_color)

    def render(self):
        pass

    def get_render(self):
        self.drawer = self.get_drawer()
        self.render()
        return self.drawer

    def update(self, key):
        return key


class Widget(WidgetBase):
    def __init__(self, size: Size = None, pos: Point = Point(0, 0), **kwargs):
        super().__init__(size, pos, **kwargs)
        # defaults
        self.is_disabled = False
        self.is_selected = False
        self.selectable = True

    def run_callbacks(self):
        pass

    def on_enter(self):
        return self.run_callbacks()

    def on_interrupt(self):
        super().on_interrupt()
        self.is_disabled = True

    def update(self, key):
        key = super().update(key)
        if key == Key.ENTER:
            self.on_enter()
            return
        return key


class WidgetLabel(Widget):
    def __init__(self,
                 text,
                 centered: bool = True,
                 wordwrap: bool = True,
                 size: Size = None,
                 pos: Point = Point(0, 0),
                 **kwargs
                 ):

        super().__init__(size, pos, **kwargs)

        if isinstance(text, str):
            self.text = CStr(text, forecolor=(220, 220, 220))
        elif isinstance(text, CStr) or isinstance(text, CStrList):
            self.text = text
        else:
            raise NotImplementedError()

        self.size.h = (len(text) // (self.size.w - 8)) + 1
        self.centered = centered
        self.wordwrap = wordwrap
        self.selectable = False

    def run_callbacks(self):
        pass

    def _place_text(self, text, ln_idx):
        if self.centered:
            self.drawer.center_place(text, ln_idx=ln_idx)
        else:
            self.drawer.place(text, pos=(0, ln_idx))

    def place_text(self, t, ln_idx):
        #print(t)
        if len(t) > (self.size.getWidth() - 8):
            for i in range(len(t)):
                i_ = len(t) - i - 1
                if i_ > (self.size.getWidth() - 8):
                    continue
                c = t[i_]
                if isinstance(c, str) and c == " ":
                    self._place_text(t[:i_], ln_idx)
                    return self.place_text(t[i_+1:], ln_idx + 1)
                if isinstance(c, CStr) and c.str == " ":
                    self._place_text(t[:i_], ln_idx)
                    return self.place_text(t[i_+1:], ln_idx + 1)
        else:
            self._place_text(t, ln_idx)

    def render(self):
        self.place_text(self.text, ln_idx=0)


class WidgetOptions(Widget):
    def __init__(self,
                 options: list[tuple[str, types.FunctionType]],
                 text: str = "Options:",
                 default_option: int = 0,
                 size: Size = None,
                 pos: Point = Point(0, 0),
                 **kwargs
                 ):

        super().__init__(size, pos, **kwargs)
        self.n_options = len(options)
        self.options = list(map(lambda x: x[0], options))
        self.text = text
        self.callbacks = list(map(lambda x: x[1], options))
        self.selected = default_option
        self.size.h = 1 + self.n_options

    def run_callbacks(self):
        self.result = self.callbacks[self.selected](self.parent)
        return self.result

    def update(self, key):
        key = super().update(key)
        if key == Key.UP:
            self.selected = (self.selected - 1) % self.n_options
            return
        if key == Key.DOWN:
            self.selected = (self.selected + 1) % self.n_options
            return
        return key

    def render(self):
        caption_start = 2
        __options = self.text.ljust(self.size.getWidth()-caption_start)
        self.drawer.place(CStr(__options, forecolor=(
            220, 220, 220)), pos=(caption_start, 0))
        for i, option in enumerate(self.options):
            if i == self.selected:
                option_color = (50, 220, 80)
                option_indicator = CStr("> ", forecolor=(220, 220, 220))
            else:
                option_color = (220, 220, 220)
                option_indicator = CStr("  ", forecolor=(220, 220, 220))
            __option = CStr(option, forecolor=option_color)
            self.drawer.place(option_indicator + __option,
                              pos=(caption_start, 1 + i))


class WidgetInput(Widget):
    def __init__(self,
                 text: str,
                 size: Size = None,
                 pos: Point = Point(0, 0),
                 **kwargs
                 ):
        super().__init__(size, pos, **kwargs)
        # p = 1
        self.size.h = 1
        self.text = text
        self.height = 1
        self.input = ""
        self.pointer = 0
        self.view = 0
        self.show_pointer = False

        # characters that this dialouge will listen to
        self.allowed_characters = default(kwargs, "allowed_chars", "")
        

        ## keys
        ctrl_keys = [
            Key.CTRL_A,
            Key.CTRL_B,
            Key.CTRL_C,
            Key.CTRL_D,
            Key.CTRL_E,
            Key.CTRL_F,
            Key.CTRL_G,
            Key.CTRL_H,
            Key.CTRL_I,
            Key.CTRL_J,
            Key.CTRL_K,
            Key.CTRL_L,
            Key.CTRL_M,
            Key.CTRL_N,
            Key.CTRL_O,
            Key.CTRL_P,
            Key.CTRL_Q,
            Key.CTRL_R,
            Key.CTRL_S,
            Key.CTRL_T,
            Key.CTRL_U,
            Key.CTRL_V,
            Key.CTRL_W,
            Key.CTRL_X,
            Key.CTRL_Y,
            Key.CTRL_Z,
        ]
        
        funcion_keys = [
            Key.F1,
            Key.F2,
            Key.F3,
            Key.F4,
            Key.F5,
            Key.F6,
            Key.F7,
            Key.F8,
            Key.F9,
            Key.F10,
            Key.F11,
            Key.F12,
        ]
        ignored_keys = [
            Key.CR,
            Key.DOWN,
            Key.UP,
            Key.LEFT,
            Key.RIGHT,
            Key.END,
            Key.HOME,
            Key.ESC,
            Key.ESC_2,
            Key.ENTER,
            Key.ENTER_2,
            Key.INSERT,
            Key.LF,
            Key.PAGE_DOWN, 
            Key.PAGE_UP, 
            Key.SUPR,
        ]

        self.__ignored_keys = ctrl_keys + funcion_keys + ignored_keys

    def run_callbacks(self):
        super().run_callbacks()
        self.parent.close()
        self.result = self.input
        return self.result

    def reset(self):
        self.input = ""
        self.pointer = 0
        self.view = 0

    def get_max_length(self):
        p = 2
        max_input_len = self.size.getWidth()-(p * 2)-len(self.text)
        return max_input_len

    def update(self, key):
        key = super().update(key)
        if key is None:
            return

        max_input_len = self.get_max_length()
        # capture and place characters
        if key != "UPDATE" and key not in self.__ignored_keys:
            self.input = self.input[:self.pointer] + \
                key + self.input[self.pointer:]
            self.pointer += 1
            if (len(self.input)-self.view) > (max_input_len-1):
                self.view += 1
            return

        # backspace
        if key == Key.BACKSPACE:
            if self.pointer != 0:
                self.input = self.input[:self.pointer -
                                        1] + self.input[self.pointer:]
            else:
                self.input = self.input[self.pointer:]
            self.pointer = max(self.pointer - 1, 0)
            if self.pointer-self.view < 0:
                self.view = max(self.view - max_input_len, 0)
            return

        # delete
        if key == Key.DELETE:
            self.input = self.input[:self.pointer] + \
                self.input[self.pointer+1:]
            # self.pointer = max(self.pointer - 1, 0)
            return

        # left arrow
        if key == Key.LEFT:
            self.pointer = max(self.pointer - 1, 0)
            if self.pointer-self.view < 0:
                self.view = max(self.view - 1, 0)
            return

        # right arrow
        if key == Key.RIGHT:
            self.pointer = min(self.pointer + 1, len(self.input))
            if self.view == (self.pointer-max_input_len):
                self.view = min(self.view + 1, len(self.input))
            return

        if key == "UPDATE":
            self.show_pointer = (self.parent.tick % 20) >= 12
        return key

    def format_textbox(self):
        text, inp, pointer, view = self.text, self.input, self.pointer, self.view
        max_input_len = self.get_max_length()
        # get the input text with fixed length plus one extra space for the pointer
        if view+max_input_len-1 < len(inp):
            inp_ = inp[view:view+max_input_len]
        else:
            inp_ = inp[view:view+max_input_len-1] + " "

        if not self.is_disabled and self.show_pointer or inp_[pointer-view] != " ":
            pointer_str = CStr(
                inp_[pointer-view], backcolor=(220, 220, 220), forecolor=(20, 20, 20))
        else:
            pointer_str = CStr(inp_[pointer-view])
        # insert the pointer into input text at the correct position
        input_plus_pointer = CStr(
            inp_[:pointer-view]) + pointer_str + CStr(inp_[pointer-view+1:])

        text_c = (50, 200, 50)

        fstr = CStr(text, forecolor=text_c)

        back_exists = CStr("<", forecolor=text_c)
        forward_exists = CStr(">", forecolor=text_c)

        fstr += (back_exists if view != 0 else " ")
        fstr += input_plus_pointer + " " * \
            (max_input_len - len(input_plus_pointer))
        fstr += (forward_exists if view+max_input_len-1 < len(inp) else " ")

        return fstr

    def render(self):
        p = 1
        if self.parent.background_color is not None:
            text_shadow = tuple(map(lambda x: x * 0.8, list(self.parent.background_color)))
            self.drawer.place(CStr(" " * (self.size.getWidth() - (p * 2) - len(self.text)), backcolor=text_shadow), pos=(p + len(self.text), 0))
        self.drawer.place(self.format_textbox(), pos=(p, 0))
        #self.drawer.write_text(str(text_shadow), pos=(0, 0))




import os
import re
import string
from enum import Enum
from sys import platform, stdin, stdout

ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')

def InitAnsi():
    # call os.system("") to active ansi colors on the windows terminal
    os.system("")

# write and flash on stdout
f = stdout

def write(data): f.write(data)
def writeln(data): f.write(data+"\n")
def read(): stdin.read()
def readln(): stdin.readline()
def flush(): f.flush()

DEBUG = False
if DEBUG:
    f = open("output.ansi", "w", encoding="utf-8")
    write = lambda x: f.write(x)
    writeln = lambda x: f.write(x + "\n")
    flush = lambda: f.flush()

# ansi colors
class ForeColors():
    Black          = "\x1B[0;30m"
    Red            = "\x1B[0;31m"
    Green          = "\x1B[0;32m"
    Yellow         = "\x1B[0;33m"
    Blue           = "\x1B[0;34m"
    Magenta        = "\x1B[0;35m"
    Cyan           = "\x1B[0;36m"
    White          = "\x1B[0;37m"

    Default        = "\x1B[0;39m"

    Dark_Gray      = "\x1B[1;30m"
    Light_Red      = "\x1B[1;31m"
    Light_Green    = "\x1B[1;32m"
    Light_Yellow   = "\x1B[1;33m"
    Light_Blue     = "\x1B[1;34m"
    Light_Magenta  = "\x1B[1;35m"
    Light_Cyan     = "\x1B[1;36m"
    Dimm_White     = "\x1B[1;37m"

    Light_Black = Dark_Gray
    Light_White = Dimm_White

class BackColors:
    Black          = "\x1B[0;40m"
    Red            = "\x1B[0;41m"
    Green          = "\x1B[0;42m"
    Yellow         = "\x1B[0;43m"
    Blue           = "\x1B[0;44m"
    Magenta        = "\x1B[0;45m"
    Cyan           = "\x1B[0;46m"
    White          = "\x1B[0;47m"

    Default        = "\x1B[0;49m"

    Dark_Gray      = "\x1B[1;40m"
    Light_Red      = "\x1B[1;41m"
    Light_Green    = "\x1B[1;42m"
    Light_Yellow   = "\x1B[1;43m"
    Light_Blue     = "\x1B[1;44m"
    Light_Magenta  = "\x1B[1;45m"
    Light_Cyan     = "\x1B[1;46m"
    Dimm_White     = "\x1B[1;47m"

    Light_Black = Dark_Gray
    Light_Gray = Dimm_White


class ColorType(Enum):
    C816COLOR = 1
    LEGACY = 2
    TRUECOLOR = 3

class AnsiColor():
    def __init__(self, rgb: tuple[int, int, int], foreground: bool = True):
        self.rgb = rgb
        self.foreground = foreground

    def __str__(self) -> str:
        return str(self.rgb)

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        return (self.rgb == other.rgb) and self.foreground == other.foreground

    # def __add__(self, other) -> str:
    #    return self.__str__() + other.__str__()

    @staticmethod
    def get_c816color_from_rgb(rgb: tuple[int, int, int], foreground: bool) -> str:
        # TODO: implement
        raise NotImplementedError()

    @staticmethod
    def get_legacycolor_from_rgb(rgb: tuple[int, int, int], foreground: bool) -> str:
        if foreground:
            return FRgbAnsi(*rgb)
        else:
            return BRgbAnsi(*rgb)

    @staticmethod
    def get_truecolor_from_rgb(rgb: tuple[int, int, int], foreground: bool) -> str:
        if foreground:
            return FRgbAnsiTC(*rgb)
        else:
            return BRgbAnsiTC(*rgb)

    def get(self, ctype: ColorType = None) -> str:
        if ctype is None:
            ctype = ColorType.LEGACY

        if ctype == ColorType.C816COLOR:
            return self.get_c816color_from_rgb(self.rgb, self.foreground)
        elif ctype == ColorType.LEGACY:
            return self.get_legacycolor_from_rgb(self.rgb, self.foreground)
        elif ctype == ColorType.TRUECOLOR:
            return self.get_truecolor_from_rgb(self.rgb, self.foreground)

def remove_ansi(text):
    return ansi_escape.sub('', text)

def round_if_float(c):
    return round(c) if isinstance(c, float) else c

def FRgbAnsiTC(R: int, G: int, B: int) -> string:
    R = round_if_float(R)
    G = round_if_float(G)
    B = round_if_float(B)
    return (f"\x1b[38;2;{R};{G};{B}m")

def BRgbAnsiTC(R: int, G: int, B: int) -> string:
    R = round_if_float(R)
    G = round_if_float(G)
    B = round_if_float(B)
    return (f"\x1b[48;2;{R};{G};{B}m")

def calc_legacy_colorbit(c):
    #legacy color ranges from about 48 - 236 (changes in different environments)
    k = (1 / 188) * 5
    r = int(max(0,(c-48)) * k)
    return r

def calc_legacy_colorcode(R: int, G: int, B: int) -> int:
    r = calc_legacy_colorbit(R)
    g = calc_legacy_colorbit(G)
    b = calc_legacy_colorbit(B)
    color_code = r * 36 + g * 6 + b + 16
    #print(color_code)
    return color_code

def FRgbAnsi(R: int, G: int, B: int) -> string:
    color_code = calc_legacy_colorcode(R, G, B)
    return (f"\x1b[38;5;{color_code}m")

def BRgbAnsi(R: int, G: int, B: int) -> string:
    color_code = calc_legacy_colorcode(R, G, B)
    return (f"\x1b[48;5;{color_code}m")

def bold(text):
    return "\x1b[1m" + text + "\x1b[22m"


def italic(text):
    return "\x1b[3m" + text + "\x1b[23m"


def underline(text):
    return "\x1b[4m" + text + "\x1b[24m"

# might not be supprted in some environments
def blinking(text):
    return "\x1b[5m" + text + "\x1b[25m"


CR = "\x1b[0m"


def go_up(n_lines: int) -> str:
    return write(f"\033[{n_lines}F")

# Legacy go up
def go_up_(n_lines: int = 1) -> str:
    return write("\033M" * n_lines)

# clear console


def clear():
    if platform.startswith(("linux", "darwin", "freebsd", "openbsd", "cygwin")):
        os.system("clear")
    elif platform == "win32":
        os.system("cls")
    else:
        raise NotImplementedError(f"The platform {platform} is not supported yet.")

# get terminal size
def get_size() -> os.terminal_size:
    return os.get_terminal_size()


import pybud.ansi as ansi
from pybud.ansi import ColorType
from pybud.deftypes import Point, Size
from pybud.drawing import ColoredString as CStr
from pybud.gui.dialog import AutoDialog, DialogBase
from pybud.gui.widgets import WidgetInput, WidgetLabel, WidgetOptions


def input_dialog(WIDTH=76):
    """ the main dialog """

    title = CStr("[") + CStr(" " + "PyBUD: GUI Beauty" +
                             " ", forecolor=(255, 0, 0)) + CStr("]")
    caption = "A python library for creating beautiful GUIs in console, with tons of diffrent components, such as Dialogs, Widgets, Drawables, color optimization, and more!"
    mydialog = AutoDialog(width=WIDTH, ctype=ColorType.TRUECOLOR, background_color = (110, 90, 250))

    mydialog.add_widget(WidgetLabel(
        title,
        size=Size(WIDTH),  # height will be owerwritten in WidgetLabel
        pos=Point(0, 1),
    ))

    mydialog.add_widget(WidgetLabel(
        caption,
        centered=True,
        size=Size(WIDTH),  # height will be owerwritten in WidgetLabel
        pos=Point(0, 2),
    ))

    mydialog.add_widget(WidgetInput(
        "text input: ",
        size=Size(WIDTH//2 - 4),  # height will be owerwritten in WidgetLabel
        pos=Point(2, 6),
    ))
    mydialog.add_widget(WidgetInput(
        "another text input: ",
        size=Size(WIDTH//2 - 4),  # height will be owerwritten in WidgetLabel
        pos=Point(WIDTH//2 + 2, 6),
    ))

    def briliant_option(self: DialogBase):
        ansi.write("Briliant!        \r")
        ansi.flush()
        return "Briliant!"

    def verycool_option(self: DialogBase):
        ansi.write("Very Cool!        \r")
        ansi.flush()
        return "Very Cool!"

    def nice_option(self: DialogBase):
        ansi.write("Nice!             \r")
        ansi.flush()
        return "Nice!"

    def awesome_option(self: DialogBase):
        ansi.write("Awesome!          \r")
        ansi.flush()
        return "Awesome!"
    # the text and callback function for each option
    options = [
        ("Nice!", nice_option),
        ("Very Cool!", verycool_option),
        ("Awesome!", awesome_option),
        ("Briliant!", briliant_option),
    ]
    mydialog.add_widget(WidgetOptions(
        options,
        size=Size(WIDTH-4),  # height will be owerwritten in WidgetOptions
        pos=Point(2, 8),
    ))
    mydialog.add_widget(WidgetLabel(
        CStr("Tip: ", forecolor=(220, 220, 0)) +
        CStr("you can use TAB or arrow keys to switch between selectable Widgets! and use ") + CStr("Ctrl + C", forecolor=(220, 220, 0)) +
        CStr(" or press enter on (InputWidget)s to exit!"),
        centered=True,
        size=Size(WIDTH - 24),  # height will be owerwritten in WidgetLabel
        pos=Point(22, 9),
        name = "LastLabel"
    ))
    return mydialog


if __name__ == "__main__":
    ansi.InitAnsi()

    mydialog = input_dialog()

    # print dialog summery
    # print(mydialog)

    # mydialog.show()
    text = mydialog.show()

    print(f"result (last update return): \"{text}\"")

    print(f"individual outputs:")
    for i, w in enumerate(mydialog.widgets):
        print(w.name + ":")
        if isinstance(w, WidgetInput):
            print(f"text written in textbox ({w.text}): ", w.input)
        elif isinstance(w, WidgetOptions):
            print("selected option id:", w.selected)
            print("selected option text:", w.options[w.selected])
        else:
            print("no output")
